def crop(matrix, row, column):
    return matrix[:row, :column]


def cropIntoRectangle(matrix, x, xLength, y, yLength):
    """
    Returns a cropped rectangle from the `matrix`
    @param matrix: matrix to crop
    @param x: xCoordinate to start crop
    @param xLength: length across first(x) axis
    @param y: yCoordinate to start crop
    @param yLength: length across second (y)axis

    @return a portion of the `matrix`
    """
    return matrix[x : x + xLength, y : y + yLength]



import sys
from functools import cache
from pathlib import Path
from typing import Callable, List, TypedDict

import toml
from typing_extensions import NotRequired

from m23.constants import DEFAULT_CPU_FRACTION_USAGE, LOG_FILES_COMBINED_FOLDER_NAME
from m23.file.log_file_combined_file import LogFileCombinedFile
from m23.processor.config_loader import (
    is_night_name_valid,
    is_valid_radii_of_extraction,
    load_configuration_with_necessary_reference_files,
)
from m23.utils import get_image_number_in_log_file_combined_file


class RenormalizeConfigProcessing(TypedDict):
    radii_of_extraction: List[int]
    cpu_fraction: NotRequired[float]


class RenormalizeConfigReference(TypedDict):
    file: Path | str
    logfile: Path | str
    color: Path | str


class RenormalizeConfigNight(TypedDict):
    path: Path | str
    first_logfile_number: int
    last_logfile_number: int
    files_to_use: NotRequired[List[Path]]


class RenormalizeConfigInput(TypedDict):
    nights: List[RenormalizeConfigNight]


class RenormalizeConfig(TypedDict):
    processing: RenormalizeConfigProcessing
    reference: RenormalizeConfigReference
    input: RenormalizeConfigInput


@cache
def get_relevant_log_files_combined_files(folder: Path, start: int, end: int) -> List[Path]:
    """
    Returns the list of log files combined files in the range enclosed
    (inclusively) by start and end path.

    param: folder: Path containing log files combined
    param: start: First file number to consider
    param: end: Last file number to consider

    Note that if start and/or end number fall of the edges of the available file
    numbers, the result contains the result that goes to the edge on the
    respective side.
    """
    result = []
    for file in folder.glob("*"):
        if file.is_file() and file.suffix == ".txt":
            if start <= get_image_number_in_log_file_combined_file(file) <= end:
                result.append(file)
    return result


def validate_night(night):
    path = Path(night["path"])

    first_logfile = night["first_logfile_number"]
    last_logfile = night["last_logfile_number"]

    # Validate path
    if not path.exists():
        sys.stderr.write(f"Path {path} doesn't exist\n")
        return False

    # Check if the naming convention is valid for the input night
    if not is_night_name_valid(path):
        sys.stderr.write("Naming convention is invalid\n")
        return False

    LOG_FILES_COMBINED_FOLDER = path / LOG_FILES_COMBINED_FOLDER_NAME

    if not LOG_FILES_COMBINED_FOLDER.exists():
        sys.stderr.write(f"Path {LOG_FILES_COMBINED_FOLDER} doesn't exist\n")
        return False

    if (type(first_logfile) != int or type(last_logfile) != int) and not (
        last_logfile >= first_logfile >= 1
    ):
        sys.stderr.write(
            f"Invalid logfile numbers. First: {first_logfile} Last: {last_logfile} for night {night}\n"  # noqa
        )
        return False

    # Ensure that the list of files to use from LOG_FILES_COMBINED_FOLDER isn't empty
    # based on given constraints of start and end indices
    if (
        len(
            get_relevant_log_files_combined_files(
                LOG_FILES_COMBINED_FOLDER, first_logfile, last_logfile
            )
        )
        == 0
    ):
        sys.stderr.write(f"No logfiles in range {first_logfile}-{last_logfile} \n")
        return False

    return True


def is_valid(config: RenormalizeConfig) -> bool:  # noqa
    """
    Returns whether any error can be found in renormalize config dict
    """
    # Validate radii of extraction
    if not is_valid_radii_of_extraction(config["processing"]["radii_of_extraction"]):
        return False

    # Validate reference file
    ref_file = Path(config["reference"]["file"])
    if not (ref_file.exists() and ref_file.is_file() and ref_file.suffix == ".txt"):
        sys.stderr.write("Make sure the provided reference file exits and has txt extension\n")
        return False

    # Validate logfile file
    logfile = Path(config["reference"]["logfile"])
    if not (logfile.exists() and logfile.is_file() and logfile.suffix == ".txt"):
        sys.stderr.write("Make sure the provided logfile file exits and has txt extension\n")
        return False

    # Make sure that the logfile combined reference file has
    # all radii of extraction data
    available_radii = LogFileCombinedFile(logfile).get_star_data(1).radii_adu.keys()
    for i in config["processing"]["radii_of_extraction"]:
        if i not in available_radii:
            sys.stderr.write(
                f"Radius {i} ADU data not present in provided logfile combined file. \n"
            )
            return False

    if cpu_fraction := config["processing"].get("cpu_fraction"):
        if not 0 <= cpu_fraction <= 1:
            sys.stderr.write(f"CPU fraction has to be between 0 and 1. Received {cpu_fraction} \n")
            return False

    color_ref_file = Path(config["reference"]["color"])
    if not (
        color_ref_file.exists() and color_ref_file.is_file() and color_ref_file.suffix == ".txt"
    ):
        sys.stderr.write(
            "Make sure the provided color reference file exits and has txt extension\n"
        )
        return False

    # Validate each night
    for night in config["input"]["nights"]:
        is_valid = validate_night(night)
        if not is_valid:
            sys.stderr.write(f"Invalid night {night}\n")
            return False

        return True  # No errors detected


def sanity_check(config: RenormalizeConfig):
    """
    Performs any sanity check in the renormalization config
    """
    return config


def create_enhanced_config(config: RenormalizeConfig):
    """
    Creates an enhanced version of the renormalize config by converting str types to Path objects
    and other possible sanitation/optimizations
    """
    if type(config["reference"]["file"]) == str:
        config["reference"]["file"] = Path(config["reference"]["file"])
    if type(config["reference"]["color"]) == str:
        config["reference"]["color"] = Path(config["reference"]["color"])

    for night in config["input"]["nights"]:
        if type(night["path"]) == str:
            night["path"] = Path(night["path"])
        # Set the list of log files to use for normalization based on the
        # provided start, end indices
        night["files_to_use"] = get_relevant_log_files_combined_files(
            night["path"] / LOG_FILES_COMBINED_FOLDER_NAME,
            night["first_logfile_number"],
            night["last_logfile_number"],
        )

    # Remove duplicates in radii of extraction
    radii = config["processing"]["radii_of_extraction"]
    config["processing"]["radii_of_extraction"] = list(set(radii))

    if config["processing"].get("cpu_fraction", None) is None:
        config["processing"]["cpu_fraction"] = DEFAULT_CPU_FRACTION_USAGE

    return config


def validate_renormalize_config_file(
    file_path: Path, on_success: Callable[[RenormalizeConfig], None]
) -> None:
    """
    This method reads data processing configuration from the file path
    provided and calls the unary function on_success if the configuration
    file is valid with the configuration dictionary (Note, *not* config file).
    """
    if not file_path.exists() or not file_path.exists():
        raise FileNotFoundError("Cannot find configuration file")
    configuration = toml.load(file_path)
    load_configuration_with_necessary_reference_files(configuration, pop=("image",))
    match configuration:
        case {
            "processing": {"radii_of_extraction": list(_)},
            "input": {"nights": list(_)},
            "reference": {"file": str(_), "color": str(_), "logfile": str(_)},
        } as renormalize_config if is_valid(renormalize_config):
            on_success(sanity_check(create_enhanced_config(renormalize_config)))
        case _:
            sys.stderr.write("Stopping\n")


if __name__ == "__main__":
    validate_renormalize_config_file(Path("1.toml"), on_success=lambda *args: print("Success"))



from pathlib import Path

from m23 import REFERENCE_DIR


def get_reference_files_dict():
    ref = {
        "color": str(Path(REFERENCE_DIR) / "MeanRI100.txt"),
        "file": str(Path(REFERENCE_DIR) / "ref_with_new_stars.txt"),
        "image": str(Path(REFERENCE_DIR) / "m23_3.5_071.fit"),
        "logfile": str(Path(REFERENCE_DIR) / "08-05-03_m23_3.5-071.txt"),
    }
    return ref



class ConfigException(Exception):
    pass


class InvalidDatetimeInConfig(ConfigException):
    pass


class CouldNotAlignException(Exception):
    pass


class InternightException(Exception):
    pass



import math
import warnings
from functools import cache
from typing import Tuple

import numpy as np


class SkyBgCalculator:
    # Something that always confused me is that position (x, y) in the reference
    # can be accessed from the image_data with image_data[y][x]

    @classmethod
    def get_box_number(cls, x: float, y: float) -> Tuple[int, int, int]:
        # We find out which sky background box the given co-ordinate falls under
        box_row, box_col = y // 128, x // 128
        small_box = (y - box_row * 128) // 32
        return box_row, box_col, small_box

    @classmethod
    def star_positions_circle_matrix(cls, x, y, radius):
        size = 2 * radius + 1
        matrix = [[1 for _ in range(size)] for _ in range(size)]
        matrix = np.multiply(matrix, circleMatrix(radius))
        matrix = matrix.astype("object")
        for row in range(-radius, radius + 1):
            for col in range(-radius, radius + 1):
                if matrix[row + radius][col + radius] != 0:
                    matrix[row + radius][col + radius] = (x + col, y + row)
        return matrix

    @classmethod
    def star_positions(cls, x, y, radius):
        matrix = cls.star_positions_circle_matrix(x, y, radius)
        return matrix.ravel()[np.flatnonzero(matrix)]

    def __init__(self, image_data) -> None:
        # This storage cache has to be image specific
        self.storage = {}
        self.image_data = image_data

    def get_image_data_at_box(self, box: Tuple[int, int, int]):
        row, col, small_box_number = [int(x) for x in box]
        start_row = row * 128
        start_col = col * 128
        return self.image_data[start_row : start_row + 128, start_col : start_col + 128][
            32 * small_box_number : 32 * (small_box_number + 1)
        ]

    def calculate_bg_at_position(self, x: float, y: float) -> float:
        """
        Returns the sky background at the given pixel position
        """
        # We divide the 1024*1024 image_data into boxes of size 128, 128. We further
        # strip the 128 sized square into four thin strips along the row then
        # calculate the sky background in that 32 * 128 (longer columns) by plotting
        # a 2nd degree polyfit to the values in that rectangular strip (only 10%
        # values from 40th to 50th percentile) as a function of their column number

        # This means that there will be a total of 8 big boxes across rows, 8 across
        # columns. In each of these 8 boxes, there will be 4 smaller rectangular
        # boxes.

        bg_box_number = self.get_box_number(x, y)
        if self.storage.get(bg_box_number) is None:
            bg_box = self.get_image_data_at_box(bg_box_number)
            list_of_x_position_and_adus = []
            for row in range(32):
                for col in range(128):
                    adu = bg_box[row][col]
                    if adu != 0:
                        list_of_x_position_and_adus.append((col, adu))

            # Sort list by ADU
            list_of_x_position_and_adus.sort(key=lambda x: x[1])
            # We now only keep 40%-55% percentile to do the fitting as anything
            # higher than that would most probably be stars, and lowers might be
            # black values
            length = len(list_of_x_position_and_adus)
            centered_array = list_of_x_position_and_adus[
                int(0.4 * length) : int(0.55 * length) + 1
            ]

            if len(centered_array) == 0:
                self.storage[bg_box_number] = lambda *args: 0
            else:
                x_to_plot, y_to_plot = zip(*centered_array)
                with warnings.catch_warnings():
                    warnings.filterwarnings("error")
                    try:
                        self.storage[bg_box_number] = np.poly1d(
                            np.polyfit(x=x_to_plot, y=y_to_plot, deg=2)
                        )
                    except np.linalg.LinAlgError:
                        # Use average in case of error. Is writing nan better so
                        # that we can ignore the star altogether?
                        # self.storage[bg_box_number] = lambda *args: np.mean(y_to_plot)
                        self.storage[bg_box_number] = lambda *args: np.nan
                    except np.RankWarning:
                        self.storage[bg_box_number] = lambda *args: np.nan
                    except RuntimeWarning:
                        self.storage[bg_box_number] = lambda *args: np.nan

        x_position_within_big_box = x - bg_box_number[1] * 128
        # The background at the given pixel given by the polynomial fit at the small
        # box Since the items stored in the storage_dict are those fit functions at
        # the respective boxes, we get the bg value by applying that function
        return self.storage[bg_box_number](x_position_within_big_box)

    def get_star_average_bg_per_pixel(self, x, y, radius):
        positions = self.star_positions(x, y, radius)
        adu_bg_sum = 0
        for position in positions:
            adu_bg_sum += self.calculate_bg_at_position(*position)
        return adu_bg_sum / len(positions)


@cache
def circleMatrix(radius):
    lengthOfSquare = radius * 2 + 1
    myMatrix = np.zeros(lengthOfSquare * lengthOfSquare).reshape(lengthOfSquare, lengthOfSquare)
    for row in range(-radius, radius + 1):
        for col in range(-radius, radius + 1):
            if math.ceil(math.sqrt((row) ** 2 + (col) ** 2)) <= radius:
                myMatrix[row + radius][col + radius] = 1
    return myMatrix



import logging
import math
from datetime import date
from pathlib import Path
from typing import Iterable, List, Tuple, TypedDict

import numpy as np
from typing_extensions import NotRequired

from m23.constants import INTRA_NIGHT_IMPACT_THRESHOLD_PIXELS
from m23.file.flux_log_combined_file import FluxLogCombinedFile
from m23.file.log_file_combined_file import LogFileCombinedFile
from m23.file.normfactor_file import NormfactorFile
from m23.file.reference_log_file import ReferenceLogFile
from m23.utils import half_round_up_to_int

from .get_line import get_star_to_ignore_bit_vector


class NormalizationTechniques:
    # Normalization linear means that we normalize with respect to 20% 40% 60%
    # and 80% image of the night
    LINEAR = "LINEAR"
    # Elevation means that we normalize to some images (1-4) from within a night
    # that have the same cluster angle (within 1 degree)
    ELEVATION = "ELEVATION"


class IntranightNormalizationResult(TypedDict):
    normfactors: Iterable[float]
    normalized_cluster_angle: NotRequired[float]


def normalize_log_files(  # noqa
    reference_log_file: ReferenceLogFile,
    log_files_to_normalize: List[LogFileCombinedFile],
    output_folder: Path,
    radius: int,
    img_duration: float,
    night_date: date,
) -> IntranightNormalizationResult:
    """
    This function normalizes (intra night *not* inter night) the
    LogFilesCombined files provided.  Note that the normalization **isn't** done
    with respect to the data in the reference image but with respect to some
    sample images take throughout the night.

    Note that this code assumes that the all stars in the log files are
    available in reference log file and no more or less.

    @return: iterable of normfactors for the images
    """

    # We are sorting the log files so that we know what's the first logfile
    # we are using and what's the last. This data is needed written in header
    # of all flux log combined files that we create
    logger = logging.getLogger("LOGGER_" + str(night_date))

    log_files_to_normalize.sort(key=lambda log_file: log_file.img_number())
    no_of_files = len(log_files_to_normalize)

    indices_to_normalize_to = get_indices_to_normalize_to(
        log_files_to_normalize, NormalizationTechniques.ELEVATION, night_date
    )
    log_file_names = list(map(lambda x: x.path().name, log_files_to_normalize))
    log_files_to_normalize_to_str = np.array(log_file_names).take(indices_to_normalize_to)
    logger.info(f"Logfiles to normalize wit respect to {log_files_to_normalize_to_str}")

    array_of_logfiles_of_array_of_adus = []  # This is an array of arrays

    # This holds the normalization factor for each log_file to use
    all_norm_factors = []

    for file_index, log_file in enumerate(log_files_to_normalize):
        adu_of_current_log_file = log_files_to_normalize[file_index].get_adu(radius)

        # Perform linear fits, cropping in 12 pixels from stars closest to the
        # four corners creating a quadrilateral region, and excluding stars
        # outside of this area
        stars_to_ignore_bit_vector = get_star_to_ignore_bit_vector(log_file, radius)

        # Mask out stars with center more than 1 pixel away from those in the
        # ref file also mask if the star is outside the 12px box around the
        # image
        for star_index in range(len(log_file)):
            # Note not to be confused between logfile and reference file here we
            # call logfile combined that's to be normalized as logfile and the
            # reference file that we use to look up the standard positions of
            # star as reference file
            star_no = star_index + 1

            # Mark the adu of the star as 0 if that's to be ignored
            if stars_to_ignore_bit_vector[star_index] == 0:
                adu_of_current_log_file[star_index] = 0
                # We go to the next star in the for loop as we already know ADU
                # for this star for this image
                continue

            star_data_in_log_file = log_file.get_star_data(star_no)
            if (
                not all([x > 0 for x in star_data_in_log_file.radii_adu.values()])
                or star_data_in_log_file.sky_adu <= 0
            ):
                adu_of_current_log_file[
                    star_index
                ] = 0  # Ignore this star for normfactor calc of this logfile
                continue

            star_x_reffile, star_y_reffile = reference_log_file.get_star_xy(star_no)
            star_x_position, star_y_position = (
                star_data_in_log_file.x,
                star_data_in_log_file.y,
            )

            if (
                math.sqrt(
                    (star_x_reffile - star_x_position) ** 2
                    + (star_y_reffile - star_y_position) ** 2
                )
                > INTRA_NIGHT_IMPACT_THRESHOLD_PIXELS
            ):
                adu_of_current_log_file[star_index] = 0

        array_of_logfiles_of_array_of_adus.append(adu_of_current_log_file)

    # Now for each log file we calculate its normfactor For each logfile, its
    # normfactor is the median of normfactors of the stars in that image.  For a
    # star, its normfactor in a logfile is sum of adus in reference log files
    # divided by 4 * its adu. Note that reference log files mean the 4 sample
    # log files taken from within the night.
    reference_log_files = []
    for index, log_file in enumerate(array_of_logfiles_of_array_of_adus):
        if index in indices_to_normalize_to:
            reference_log_files.append(log_file)

    array_of_logfiles_of_array_of_adus = np.array(
        array_of_logfiles_of_array_of_adus
    )  # Convert to numpy array
    reference_log_files = np.array(reference_log_files)  # Convert to numpy array
    for file_index, log_file in enumerate(array_of_logfiles_of_array_of_adus):
        no_of_stars = len(log_file)
        norm_factor_for_stars = []
        for star_index in range(no_of_stars):
            star_adu_in_reference_log_files = []
            for ref_log_file in reference_log_files:
                star_adu_in_reference_log_files.append(ref_log_file[star_index])
            star_adu = log_file[star_index]
            if all([value > 0 for value in star_adu_in_reference_log_files]) and star_adu > 0:
                normfactor = sum(star_adu_in_reference_log_files) / (
                    len(indices_to_normalize_to) * star_adu
                )
            else:
                normfactor = 0
            norm_factor_for_stars.append(normfactor)
        good_scale_factors = [x for x in norm_factor_for_stars if 0 < x <= 5]
        norm_factor_for_logfile = np.median(good_scale_factors)
        all_norm_factors.append(norm_factor_for_logfile)
        array_of_logfiles_of_array_of_adus[file_index] = (
            norm_factor_for_logfile * array_of_logfiles_of_array_of_adus[file_index]
        )

    # Save normfactors
    normfactors_file_name = NormfactorFile.generate_file_name(night_date, img_duration)
    normfactor_file = NormfactorFile(output_folder / normfactors_file_name)
    normfactor_file.create_file(all_norm_factors)

    # Save the normalized data for each star
    noOfStars = len(array_of_logfiles_of_array_of_adus[0])
    for star_index in range(noOfStars):
        star_no = star_index + 1
        star_adu_data = [
            array_of_logfiles_of_array_of_adus[file_index][star_index]
            for file_index in range(no_of_files)
        ]
        # Turn all star_adu_data that's negative to 0
        star_adu_data = [current_data if current_data > 0 else 0 for current_data in star_adu_data]

        # We now create flux log combined file
        flux_log_combined_file_name = FluxLogCombinedFile.generate_file_name(
            night_date, star_no, img_duration
        )
        flux_log_combined_file = FluxLogCombinedFile(output_folder / flux_log_combined_file_name)

        fist_log_file_number = log_files_to_normalize[0].img_number()
        last_log_file_number = log_files_to_normalize[-1].img_number()

        x_positions = []
        y_positions = []
        date_times = []

        for lf in log_files_to_normalize:
            star_data = lf.get_star_data(star_no)
            x_positions.append(star_data.x)
            y_positions.append(star_data.y)
            date_times.append(lf.datetime())

        flux_log_combined_file.create_file(
            star_adu_data,
            fist_log_file_number,
            last_log_file_number,
            x_positions,
            y_positions,
            all_norm_factors,
            date_times,
            reference_log_file,
        )

    return {
        "normfactors": all_norm_factors,
        "normalized_cluster_angle": get_cluster_angle_to_normalize_by_log_files(
            log_files_to_normalize
        ),
    }


def get_cluster_angle_to_normalize_by_log_files(log_files) -> float:
    angles_and_file_numbers = []
    for logfile in log_files:
        angles_and_file_numbers.append((logfile.get_cluster_angle(), logfile.img_number()))
    angle = get_cluster_angle_to_normalize_to(angles_and_file_numbers)
    return angle


def get_indices_to_normalize_to(log_files, technique, night_date):
    logger = logging.getLogger("LOGGER_" + str(night_date))
    if technique == NormalizationTechniques.LINEAR:
        # Normalization is done with reference to images 20%, 40%, 60% and 80%
        # throughout night The indices here are the index of the images from the
        # night to which to normalize.
        return np.linspace(0, len(log_files), 6, dtype="int")[1:-1]
    elif technique == NormalizationTechniques.ELEVATION:
        angle = get_cluster_angle_to_normalize_by_log_files(log_files)
        logger.info(f"Cluster angle to normalized to {angle}")
        indices = []
        for index, logfile in enumerate(log_files):
            if half_round_up_to_int(logfile.get_cluster_angle()) == angle:
                indices.append(index)
        if len(indices) < 4:
            return indices
        else:
            indices_of_indices = np.linspace(0, len(indices), 6, dtype="int")[1:-1]
            return np.array(indices).take(indices_of_indices)
    else:
        raise Exception(f"Intranight {technique} not recognized")


def get_cluster_angle_to_normalize_to(
    angles_and_file_number_exact: Iterable[Tuple[float, int]]
) -> int:
    """
    @param: angles_and_file_number: An iterable of tuple of angle and image number
    @returns: cluster angle to normalize to

    Note that we want to choose a cluster angle that's present the most in the
    night.  If there are more than one angles, we choose the one that spread out
    as far as possible into the night. If there's are more than one contenders
    present, we choose the one that's closest to the middle of the night.
    """

    # Rounded version
    angles_and_file_number = list(
        map(lambda x: (half_round_up_to_int(x[0]), x[1]), angles_and_file_number_exact)
    )

    angles, file_numbers = zip(*angles_and_file_number)

    file_numbers = sorted(file_numbers)
    first_file_number = file_numbers[0]
    last_file_number = file_numbers[-1]

    # Create a dictionary that holds count of each angle
    angles_count_dict = {angle: 0 for angle in set(angles)}
    for angle in angles:
        angles_count_dict[angle] += 1

    angles_count_sorted = sorted(angles_count_dict.items(), key=lambda x: x[1], reverse=True)
    max_count = angles_count_sorted[0][1]

    # Angles have have max count (there could be multiple) are stored as candidates
    candidates_angles = list(
        map(lambda x: x[0], (filter(lambda x: x[1] == max_count, angles_count_sorted)))
    )

    if len(candidates_angles) == 1:
        return candidates_angles[0]

    # Find a candidate that has the most spread
    candidates_spread = {}
    candidates_first_file_number = {}
    for candidate in candidates_angles:
        by_candidate = sorted(
            filter(lambda x: x[0] == candidate, angles_and_file_number), key=lambda x: x[1]
        )
        first_file_by_candidate = by_candidate[0][1]
        candidates_first_file_number[candidate] = first_file_by_candidate
        last_file_by_candidate = by_candidate[-1][1]
        candidates_spread[candidate] = last_file_by_candidate - first_file_by_candidate

    # Find the candidate with most spread
    candidates_spread_sorted = sorted(candidates_spread.items(), key=lambda x: x[1], reverse=True)
    max_spread = candidates_spread_sorted[0][1]

    candidates_with_max_spread = list(
        map(lambda x: x[0], filter(lambda x: x[1] == max_spread, candidates_spread_sorted))
    )

    if len(candidates_with_max_spread) == 1:
        return candidates_with_max_spread[0]

    # Return the candidate whose first image is the closest to the middle of the night
    return sorted(
        candidates_with_max_spread,
        key=lambda x: abs(
            0.5 - (candidates_first_file_number[x] / (last_file_number - first_file_number))
        ),
    )[0]



import datetime
from typing import Tuple

import numpy as np


def angle_of_elevation(date: datetime.datetime, object="M23") -> Tuple[float, float]:
    """
    Returns the angle of elevation of the interested sky object with its
    uncertainty The date and time is given in UTC.

    param: date: datetime of observation
    param: object: the name of the object you're interested. Default: M23

    return: tuple of the angle of elevation and the uncertainty of measurement
    """
    yr, mo, dy = date.year, date.month, date.day
    hour, minute, sec = date.hour, date.minute, date.second

    long = 91.79  # This is the longitude of Decorah where we do our observations
    Lat = 43.3017  # This is the latitude of Decorah where we do our observations

    if object == "M23":
        RA = (17 / 24 + 58 / (60 * 24) + 16 / (3600 * 24)) * 360
        DEC = -(19 + 1 / 60 + 7 / 3600)
    elif object == "NGC2286":
        RA = (6 / 24 + 48 / (60 * 24) + 44 / (3600 * 24)) * 360
        DEC = -(3 + 10 / 60 + 8 / 3600)
    elif object == "NGC129":
        RA = (0 / 24 + 31 / (60 * 24) + 14 / (3600 * 24)) * 360
        DEC = 60 + 20 / 60 + 12 / 3600
    elif object == "NGC7380":
        RA = (22 / 24 + 48 / (60 * 24) + 14 / (3600 * 24)) * 360
        DEC = 58 + 14 / 60 + 51 / 3600
    else:
        raise Exception("Invalid object")

    angle_list = np.zeros(31)
    for n in range(31):
        # The following code is all done to find the local sidereal time
        minu = minute - 15 + 30 / len(angle_list) * (n + 1)
        # Varying the time splits the 30 minutes into n chunks.

        JDo = (
            367 * (yr)
            - np.floor(7 * (yr + np.floor((mo + 9) / 12)) / 4)
            + np.floor((275 * mo) / 9)
            + dy
            + 1721013.5
        )
        # Calculates the julian date
        tuti = (JDo - 2451545.0) / 36525
        # The number of julian centuries since the epoch J2000

        GSToo = (
            100.4606184
            + 36000.77005361 * tuti
            + 0.00039 * tuti**2
            - 2.6 * 10 ** (-8) * tuti**3
        ) % 360  # GST at the beginning of the dy of interest

        GST = (GSToo + 0.25068447733746215 * (hour * 60 + minu + sec / 60)) % 360

        LST = (GST - long) % 360

        # We now have the RA and DEC of our zenith at the current time. We can use
        # simple spherical trig to calculate the angular separation between the
        # cluster/object and our zenith

        zAngle = np.degrees(
            np.arccos(
                np.sin(DEC * np.pi / 180) * np.sin(Lat * np.pi / 180)
                + np.cos(DEC * np.pi / 180)
                * np.cos(Lat * np.pi / 180)
                * np.cos((LST - RA) * np.pi / 180)
            )
        )
        altAngle = 90 - zAngle
        angle_list[n] = altAngle

    angle_of_elevation_unc = np.std(angle_list)

    uncertainty = round(angle_of_elevation_unc, 1)

    if uncertainty >= 10:
        AE = np.round(np.mean(angle_list), -1)
    elif uncertainty >= 1:
        AE = np.round(np.mean(angle_list))
    elif uncertainty >= 0.1:
        AE = np.round(np.mean(angle_list), 1)
    elif uncertainty >= 0.01:
        AE = np.round(np.mean(angle_list), 2)
    else:
        AE = np.mean(angle_list)

    return (AE, uncertainty)



import math

import numpy as np


def flux_to_magnitude(flux: float, radius: int) -> float:
    if not flux > 0:
        return np.nan
    if radius == 5:
        return 23.99 - 2.5665 * math.log10(flux)
    elif radius == 4:
        return 24.176 - 2.6148 * math.log10(flux)
    elif radius == 3:
        return 23.971 - 2.5907 * math.log10(flux)
    else:
        raise ValueError(f"No formula to convert for radius {radius}")



from pathlib import Path

import numpy as np

from m23.file import is_string_float


class RIColorFile:
    header_rows = 1  # First x rows that contain header information

    def __init__(self, file_path: str) -> None:
        self.__path = Path(file_path)
        self.__data = None
        self.__is_read = False

    def _validate_file(self):
        if not self.path().exists():
            raise FileNotFoundError(f"{self.path()} not found")
        if not self.path().is_file() or self.path().suffix != ".txt":
            raise ValueError(f"{self.path()} is not a valid txt file")

    def _read(self):
        self._validate_file()
        with self.path().open() as fd:
            lines = [line.strip() for line in fd.readlines()]
            lines = lines[self.header_rows :]  # Skip the header rows
            self.__data = np.array([x for x in lines if is_string_float(x)], dtype="float")
            self.__is_read = True

    def data(self):
        if not self.__is_read:
            self._read()
        return self.__data

    def get_star_color(self, star_no: int):
        if not self.__is_read:
            self._read()
        return self.data()[star_no - 1]  # Subtract 1 because of zero indexing

    def path(self):
        return self.__path

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"R-I color file: {self.path()}"



import re
from collections import namedtuple
from datetime import date, datetime
from pathlib import Path
from typing import Dict

import numpy as np

from m23.utils import get_radius_folder_name


class ColorNormalizedFile:
    StarData = namedtuple(
        "StarRow",
        [
            "median_flux",  # Median flux for the night
            "normalized_median_flux",  # Normalized flux for the night
            "norm_factor",  # Norm factor used
            "measured_mean_r_i",  # The R-I value defined in reference file
            "used_mean_r_i",  # R-I actually used to calculate norm factor
            "attendance",  # Attendance of that star for the night
            "reference_log_adu",  # Star ADU in the reference file
        ],
    )
    Data_Dict_Type = Dict[int, StarData]

    # Class attributes
    header_rows = 6  # Specifies the first x rows that don't contain header information
    file_name_re = re.compile("(\d{4}-\d{2}-\d{2})_Normalized_(.*)\.txt")

    @classmethod
    def get_file_name(cls, night_date: date, radius_of_extraction: int) -> str:
        return f"{night_date.strftime('%Y-%m-%d')}_Normalized_{get_radius_folder_name(radius_of_extraction)}.txt"  # noqa

    def __init__(self, file_path: Path) -> None:
        self.__path = file_path
        self.__read_data = False

    def save_data(self, data_dict: Data_Dict_Type, night_date: date):
        if self.__path.is_dir() or self.__path.suffix != ".txt":
            raise ValueError(f"Given path {self.__path} is not a valid txt file")

        # Create parent directories if needed
        parent_dir = self.__path.parent
        parent_dir.mkdir(parents=True, exist_ok=True)

        # Open file in writing mode
        with self.__path.open("w") as fd:
            fd.write(f"Color-normalized Data for {night_date.strftime('%Y-%m-%d')}\n")
            fd.write("\n")
            headers = [
                "Star #",
                "Normalized Median Flux",
                "Norm Factor",
                "Measured Mean R-I",
                "Used Mean R-I",
            ]
            fd.write(
                f"{headers[0]:<8s}{headers[1]:>32s}{headers[2]:>24s}{headers[3]:>32s}{headers[4]:>32s}\n"  # noqa
            )
            for star_no in sorted(data_dict.keys()):
                star_data = data_dict[star_no]
                fd.write(
                    f"{star_no:<8d}{star_data.normalized_median_flux:>32.7f}{star_data.norm_factor:>24.7f}{star_data.measured_mean_r_i:>32.7f}{star_data.used_mean_r_i:>32.7f}\n"  # noqa
                )

    def _read(self):
        self.__read_data = True
        with self.path().open() as fd:
            lines = [line.strip() for line in fd.readlines()]
            lines = lines[3:]  # Skip the header rows
            self.__data = {}
            for line in lines:
                star_data = line.split()
                star_no = int(star_data[0])
                normalized_median_flux = float(star_data[1])
                normfactor = float(star_data[2])
                measured_mean_ri = float(star_data[3])
                used_mean_ri = float(star_data[4])
                self.__data[star_no] = self.StarData(
                    normalized_median_flux=normalized_median_flux,
                    norm_factor=normfactor,
                    measured_mean_r_i=measured_mean_ri,
                    used_mean_r_i=used_mean_ri,
                    attendance=np.nan,
                    reference_log_adu=np.nan,
                    median_flux=np.nan,
                )
        self.__read_data = True  # Marks file as read

    def is_valid_file_name(self):
        return bool(self.file_name_re.match(self.path().name))

    def path(self):
        return self.__path

    def night_date(self) -> date | None:
        """
        Returns the night date that can be inferred from the file name
        """
        if self.is_valid_file_name():
            # The first capture group contains the night date
            return datetime.strptime(
                self.file_name_re.match(self.path().name)[1],
                "%Y-%m-%d",
            ).date()

    def data(self):
        if not self.__read_data:
            self._read()
        return self.__data



import numpy as np

from m23.constants import ASSUMED_MAX_BRIGHTNESS
from m23.matrix import cropIntoRectangle
from m23.utils import customMedian

# This file is for code related to applying master calibrations (dark, flats)
#   onto raw images

# Steps:
# 1. Compute average pixel value of the master flat using a square matrix
# 2. Using CALIBRATION = (AVERAGE_FLAT/MASTER_FLAT)*(RAW_IMAGE-MASTER_DARK)
#    Note: AVERAGE_FLAT is a value, MASTER_FLAT, RAW_IMAGE and MASTER_DARK are matrices.
# TODO:
#  Remove the hot pixels
#  Something to do with saving pixel values > 2 sigma in the IDL code ???

# getCenterAverage
#
# make a square center for the master flat-field frame (size is up to the user)
# and compute the average value of that square matrix
#
# 1024 x 1024 is cropped into 175x175 square box
# so 2048 x 2048 is cropped into 350 x 350 square box
# this is a square starting from (850, 850) to (1200, 1200)


def getCenterAverage(matrix):
    squareSize = 350 if matrix.shape[0] == 2048 else 175
    xPosition = 850 if matrix.shape[0] == 2048 else 425
    yPosition = 850 if matrix.shape[0] == 2048 else 425
    center = cropIntoRectangle(matrix, xPosition, squareSize, yPosition, squareSize)
    return np.mean(center)


# applyCalibration
#
# Please note that this code is a direct implementation of steps
# mentioned in Handbook of Astronomical Image Processing by
# Richard Berry and James Burnell section 6.3 Standard Calibration
#
# parameters:
#   rawImage: image name to calibrate
#   masterDarkData: masterDark data
#   masterFlatData: masterFlat data
#   averageFlatData: average value of center square of master flat
#   fileName: file Name to save the calibrated image to (needed if saving image)
#   hotPixelsInMasterDark: array of positions of hot pixels in master dark
#   save: whether or not to save image after calibration (optional, default : False)
#
# returns
#   calibrated image


def applyCalibration(
    imageData,
    masterDarkData,
    masterFlatData,
    averageFlatData,
    hotPixelsInMasterDark,
):
    # Calibration Step:

    # Completely Ignore bias
    # if len(masterBiasData):
    # imageData = imageData - masterBiasData
    # masterDarkData = masterDarkData - masterBiasData

    subtractedRaw = imageData - masterDarkData
    flatRatio = np.array(averageFlatData / masterFlatData)
    # dtype is set to float32 for our image viewing software Astromagic, since
    # it does not support float64 We think we are not losing any significant
    # precision with this down casting

    calibratedImage = np.multiply(flatRatio, subtractedRaw, dtype="float32")

    # Unlike current IDL Code we're doing this step to calibrated
    #   image instead of the raw
    # Calculate the median and standard deviation of the raw image

    medianInRaw = customMedian(imageData)
    stdInRaw = np.std(imageData)

    # NOTE We don't know why it's 2 sigma???
    highValue = medianInRaw + 2 * stdInRaw
    lowValue = medianInRaw - 2 * stdInRaw

    # recalibrate the pixels in hot positions (which are defined by
    #   hot pixels in masterDark)
    for pixelLocation in hotPixelsInMasterDark:
        recalibrateAtHotLocation(pixelLocation, calibratedImage, highValue, lowValue)

    # This calibration formula converts low background values to very high values,
    # sometimes up to millions, whereas the maximum signal of stars is less than
    # one hundred thousand
    # This will affect our alignment star-finding algorithm,
    # so we want to set these values to 0

    calibratedImage[calibratedImage > ASSUMED_MAX_BRIGHTNESS] = 0

    # Only create file if fileName is provided
    return calibratedImage


def recalibrateAtHotLocation(location, calibratedImageData, highValue, lowValue):
    # For all hot pixel positions that aren't at edges (in the master dark)
    # Check if the pixel value in calibrated( or RAW ??? TO FIX) img is abnormally
    #   high + one of the surrounding pixels is abnormally high too,
    #   then we fit a gaussian of surrounding 10X10 pixel box
    #   and assign the gaussian's value at position [5,5] (because that's the center
    #   pixel we started with) to that pixel value.
    #  ELSE: In other words:
    #  If the pixel value is not abnormally high, or if it's abnormally high but none of
    #    its surrounding pixels is abnormally high:
    #    create a 3X3 box with our pixel at center, and take the average of 8 pixels around it

    row, col = location

    # This is the smallest surrounding
    # We use 11*11 surrounding for Gaussian
    # and 3*3 for average
    def surroundingValues():
        return [
            calibratedImageData[row - 1, col],
            calibratedImageData[row + 1, col],
            calibratedImageData[row, col - 1],
            calibratedImageData[row, col + 1],
        ]

    def needsGaussian():
        # isHigh
        isHigh = calibratedImageData[row][col].all() > highValue and any(
            [value > highValue for value in surroundingValues()]
        )
        # isLow
        isLow = calibratedImageData[row][col].all() < lowValue and any(
            [value < lowValue for value in surroundingValues()]
        )
        return isHigh or isLow

    def doGaussian():
        # Create a gaussian matrix for the surrounding matrix
        # Let the hot pixel value equal to the middle position value of the gaussian box
        # Gaussian needs to be done, for now though we're just taking average.
        # TODO
        takeAverage()

    def takeAverage():
        surroundingMatrix = calibratedImageData[row - 1 : row + 2, col - 1 : col + 2]
        surroundingSum = np.sum(surroundingMatrix)
        surroundingMatrixAverageWithoutCenter = (
            surroundingSum - calibratedImageData[row][col]
        ) / 8
        calibratedImageData[row][col] = surroundingMatrixAverageWithoutCenter

    doGaussian() if needsGaussian() else takeAverage()


# A word of caution:
#   When we need the fileName, we'll call it xxxFileName
#   and when we just need the fits data in that file, we will call
#   it xxxData
#
#   For example: masterDarkFileName if fileName for masterDark image is required
#   and masterDarkData if the data of that fit file is required!


# HEADER COMMENTS: TODO

# purpose:
#   takes a list of image data to calibrate
#   returns array of calibrated image data,


def calibrateImages(masterDarkData, masterFlatData, listOfImagesData, masterBiasData=np.array([])):
    # We save the hot pixels, which are 3 standard deviation higher than the median
    # We will save their positions (x,y)
    stdInMasterDark = np.std(masterDarkData)
    medianInMasterDark = customMedian(masterDarkData)

    # Find hot pixel positions
    hotPixelPositions = np.column_stack(
        np.where(masterDarkData > medianInMasterDark + 3 * stdInMasterDark)
    )

    # We define the edges as the outermost 5 (or 10???) pixels????
    edgeSize = 5

    # noOfRows and columns in masterdarkData
    totalRows, totalColumns = masterDarkData.shape[0], masterDarkData.shape[1]

    # Filter out the edges
    # Filter top/left
    topLeftFiltered = filter(
        lambda row_column: not (row_column[0] < edgeSize or row_column[1] < edgeSize),
        hotPixelPositions,
    )
    # Filter bottom/right & convert to tuple
    filteredHotPixelPositions = tuple(
        filter(
            lambda row_column: not (
                row_column[0] > totalRows - edgeSize or row_column[1] > totalColumns - edgeSize
            ),
            topLeftFiltered,
        )
    )

    averageFlat = (getCenterAverage(masterFlatData),)

    # print("NO OF HOT PIXEL", len(filteredHotPixelPositions))
    # We need to find the flux values of (x,y) in the calibrated images

    return [
        applyCalibration(
            imageData,
            masterDarkData,
            masterFlatData,
            averageFlat,
            hotPixelsInMasterDark=filteredHotPixelPositions,
        )
        for imageData in listOfImagesData
    ]



import logging
import os
import shutil
import sys
import traceback
from datetime import date
from pathlib import Path
from typing import Iterable, List

import multiprocess as mp
import toml
from astropy.io.fits import getdata

from m23 import __version__
from m23.calibrate.master_calibrate import makeMasterDark
from m23.charts import draw_normfactors_chart
from m23.constants import (
    ALIGNED_COMBINED_FOLDER_NAME,
    ALIGNED_FOLDER_NAME,
    CONFIG_FILE_NAME,
    FLUX_LOGS_COMBINED_FOLDER_NAME,
    INPUT_CALIBRATION_FOLDER_NAME,
    LOG_FILES_COMBINED_FOLDER_NAME,
    M23_RAW_IMAGES_FOLDER_NAME,
    MASTER_DARK_NAME,
    OUTPUT_CALIBRATION_FOLDER_NAME,
    RAW_CALIBRATED_FOLDER_NAME,
    SKY_BG_BOX_REGION_SIZE,
    SKY_BG_FOLDER_NAME,
)
from m23.exceptions import InternightException
from m23.extract import sky_bg_average_for_all_regions
from m23.file.aligned_combined_file import AlignedCombinedFile
from m23.file.alignment_stats_file import AlignmentStatsFile
from m23.file.log_file_combined_file import LogFileCombinedFile
from m23.file.raw_image_file import RawImageFile
from m23.file.reference_log_file import ReferenceLogFile
from m23.file.sky_bg_file import SkyBgFile
from m23.internight_normalize import internight_normalize
from m23.matrix import crop
from m23.norm import normalize_log_files
from m23.processor.align_combined_extract import align_combined_extract
from m23.processor.config_loader import Config, ConfigInputNight, validate_file
from m23.utils import (
    fit_data_from_fit_images,
    get_all_fit_files,
    get_darks,
    get_date_from_input_night_folder_name,
    get_log_file_name,
    get_output_folder_name_from_night_date,
    get_radius_folder_name,
    get_raw_images,
)


def normalization_helper(  # noqa
    radii_of_extraction: List[int],
    reference_log_file: ReferenceLogFile,
    log_files_to_use: List[LogFileCombinedFile],
    img_duration: float,
    night_date: date,
    color_ref_file_path: Path,
    output: Path,
    logfile_combined_reference_logfile: LogFileCombinedFile,
    is_running_as_part_of_process=False,
):
    """
    This is a normalization helper function extracted so that it can be reused
    by the renormalization script
    """

    # If running as part of process, we save flux log combined in a special
    # folder will all images from the night included which will be helpful for
    # Eclipsing binary study.
    if is_running_as_part_of_process:
        FLUX_LOGS_COMBINED_OUTPUT_FOLDER = output / (FLUX_LOGS_COMBINED_FOLDER_NAME + "(All)")
    else:
        FLUX_LOGS_COMBINED_OUTPUT_FOLDER = output / FLUX_LOGS_COMBINED_FOLDER_NAME
    FLUX_LOGS_COMBINED_OUTPUT_FOLDER.mkdir(exist_ok=True)  # Create folder if it doesn't exist
    logger = logging.getLogger("LOGGER_" + str(night_date))

    if len(log_files_to_use) < 4:
        logger.error("Less than 4 data points present. Skipping normalization.")
        return

    intranight_norm_results = {}

    for radius in radii_of_extraction:
        logger.info(f"Normalizing for radius of extraction {radius} px")
        RADIUS_FOLDER = FLUX_LOGS_COMBINED_OUTPUT_FOLDER / get_radius_folder_name(radius)
        RADIUS_FOLDER.mkdir(exist_ok=True)  # Create folder if it doesn't exist
        for file in RADIUS_FOLDER.glob("*"):
            if file.is_file():
                file.unlink()  # Remove each file in the folder
        intranight_norm_result = normalize_log_files(
            reference_log_file,
            log_files_to_use,
            RADIUS_FOLDER,
            radius,
            img_duration,
            night_date,
        )
        intranight_norm_results[radius] = intranight_norm_result

    draw_normfactors_chart(log_files_to_use, FLUX_LOGS_COMBINED_OUTPUT_FOLDER, radii_of_extraction)
    logger.info("Completed drawing normfactors chart")

    # Stop running further if running as part of process.  This is because, one
    # would usually have to run renorm with a section of a night and it is then that
    # the final color normalized files, sky background files are generated. Therefore
    # it's not necessary to generate them twice.
    if is_running_as_part_of_process:
        return

    # Generate sky bg file
    sky_bg_filename = output / SKY_BG_FOLDER_NAME / SkyBgFile.generate_file_name(night_date)

    # Internight normalization
    try:
        normfactors = internight_normalize(
            output,
            logfile_combined_reference_logfile,
            color_ref_file_path,
            radii_of_extraction,
        )
    except InternightException:
        logger.error("Returning in middle of internight normalization")
        return

    # Create folder if it doesn't exist
    sky_bg_filename.parent.mkdir(parents=True, exist_ok=True)
    color_normfactors = {
        radius: normfactors[radius]["color"].items() for radius in radii_of_extraction
    }
    brightness_normfactors = {
        radius: normfactors[radius]["brightness"].items() for radius in radii_of_extraction
    }

    color_normfactors_titles = []
    color_normfactors_values = []

    for radius in color_normfactors:
        for section, section_value in color_normfactors[radius]:
            color_normfactors_titles.append(f"norm_{radius}px_color_{section}")
            color_normfactors_values.append(section_value)

    brightness_normfactors_titles = []
    brightness_normfactors_values = []

    for radius in brightness_normfactors:
        for section, section_value in brightness_normfactors[radius]:
            brightness_normfactors_titles.append(f"norm_{radius}px_brightness{section}")
            brightness_normfactors_values.append(section_value)

    create_sky_bg_file(
        SkyBgFile(sky_bg_filename),
        log_files_to_use,
        night_date,
        color_normfactors_titles,
        color_normfactors_values,
        brightness_normfactors_titles,
        brightness_normfactors_values,
        intranight_norm_results[radii_of_extraction[0]].get("normalized_cluster_angle"),
    )


def create_sky_bg_file(
    sky_bg_file: SkyBgFile,
    log_files_to_use: Iterable[LogFileCombinedFile],
    night_date: date,
    color_normfactors_title: Iterable[str],
    color_normfactors_values: Iterable[float],
    brightness_normfactors_title: Iterable[str],
    brightness_normfactors_values: Iterable[float],
    normalized_cluster_angle: int,
):
    """
    Creates sky bg data. Note that this isn't performed right after extraction
    is that we want to re-perform it after re-normalization. If we do it as part
    of `normalization_helper` which is what both `process_night` and `renorm`
    use, we wouldn't have to do it twice.

    param: sky_bg_file: SkyBgFile object to use
    param: log_files_to_use: List of log files to use
    param: night_date: Date object of the night
    normfactors: Dictionary of normfactors for various radii of extraction

    """
    logger = logging.getLogger("LOGGER_" + str(night_date))
    logger.info("Generating sky background file")
    bg_data_of_all_images = []

    for logfile in log_files_to_use:
        date_time_of_image = logfile.datetime()
        # Here we find the corresponding aligned combined file first
        # so we can use that to calculate the sky bg data.
        aligned_combined_folder = logfile.path().parent.parent / ALIGNED_COMBINED_FOLDER_NAME
        aligned_combined_file_name = AlignedCombinedFile.generate_file_name(
            logfile.img_duration(), logfile.img_number()
        )
        aligned_combined_file = AlignedCombinedFile(
            aligned_combined_folder / aligned_combined_file_name
        )
        bg_data_of_image = sky_bg_average_for_all_regions(
            aligned_combined_file.data(), SKY_BG_BOX_REGION_SIZE
        )
        image_number = aligned_combined_file.image_number()

        # Append tuple of result
        bg_data_of_all_images.append(
            (
                date_time_of_image,
                bg_data_of_image,
                image_number,
            )
        )

    sky_bg_file.create_file(
        bg_data_of_all_images,
        color_normfactors_title,
        color_normfactors_values,
        brightness_normfactors_title,
        brightness_normfactors_values,
        log_files_to_use[0].img_number(),
        log_files_to_use[-1].img_number(),
        normalized_cluster_angle,
    )
    logger.info("Completed generating sky background file")


def process_night(night: ConfigInputNight, config: Config, output: Path, night_date: date):  # noqa
    """
    Processes a given night of data based on the settings provided in `config` dict
    """
    # Save the config file used to do the current data processing
    CONFIG_PATH = output / CONFIG_FILE_NAME
    with CONFIG_PATH.open("w+") as fd:
        toml.dump(config, fd)

    # Number of expected rows and columns in all raw images
    rows, cols = config["image"]["rows"], config["image"]["columns"]
    radii_of_extraction = config["processing"]["radii_of_extraction"]
    image_duration = config["processing"]["image_duration"]
    dark_prefix = config["processing"]["dark_prefix"]

    log_file_path = output / get_log_file_name(night_date)
    # Clear file contents if exists, so that reprocessing a night wipes out
    # contents instead of appending to it
    if log_file_path.exists():
        log_file_path.unlink()

    logger = logging.getLogger("LOGGER_" + str(night_date))
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    ch = logging.FileHandler(log_file_path)
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    # Write to std out in addition to writing to a logfile
    ch2 = logging.StreamHandler(sys.stdout)
    ch2.setFormatter(formatter)
    logger.addHandler(ch2)  # Write to stdout
    logger.info(f"Starting processing for {night_date} with m23 version: {__version__}")

    ref_file_path = config["reference"]["file"]
    color_ref_file_path = config["reference"]["color"]

    reference_log_file = ReferenceLogFile(ref_file_path)
    logfile_combined_reference_logfile = LogFileCombinedFile(config["reference"]["logfile"])

    # Define relevant input folders for the night being processed
    NIGHT_INPUT_FOLDER: Path = night["path"]
    NIGHT_INPUT_CALIBRATION_FOLDER: Path = NIGHT_INPUT_FOLDER / INPUT_CALIBRATION_FOLDER_NAME
    NIGHT_INPUT_IMAGES_FOLDER = NIGHT_INPUT_FOLDER / M23_RAW_IMAGES_FOLDER_NAME

    # Define and create relevant output folders for the night being processed
    JUST_ALIGNED_NOT_COMBINED_OUTPUT_FOLDER = output / ALIGNED_FOLDER_NAME
    CALIBRATION_OUTPUT_FOLDER = output / OUTPUT_CALIBRATION_FOLDER_NAME
    ALIGNED_COMBINED_OUTPUT_FOLDER = output / ALIGNED_COMBINED_FOLDER_NAME
    LOG_FILES_COMBINED_OUTPUT_FOLDER = output / LOG_FILES_COMBINED_FOLDER_NAME
    FLUX_LOGS_COMBINED_OUTPUT_FOLDER = output / FLUX_LOGS_COMBINED_FOLDER_NAME
    RAW_CALIBRATED_OUTPUT_FOLDER = output / RAW_CALIBRATED_FOLDER_NAME

    for folder in [
        JUST_ALIGNED_NOT_COMBINED_OUTPUT_FOLDER,
        CALIBRATION_OUTPUT_FOLDER,
        RAW_CALIBRATED_OUTPUT_FOLDER,
        ALIGNED_COMBINED_OUTPUT_FOLDER,
        LOG_FILES_COMBINED_OUTPUT_FOLDER,
        FLUX_LOGS_COMBINED_OUTPUT_FOLDER,
    ]:
        if folder.exists():
            [file.unlink() for file in folder.glob("*") if file.is_file()]  # Remove existing files
        folder.mkdir(exist_ok=True)

    # Darks
    darks = fit_data_from_fit_images(
        get_darks(NIGHT_INPUT_CALIBRATION_FOLDER, image_duration, dark_prefix)
    )
    # Ensure that image dimensions are as specified by rows and cols
    # If there's extra noise cols or rows, we crop them
    # Note this is different from the crop_region that's defined in image
    # options for process. More than crop, it's a fill that fills out the
    # vignetting ring with zero values
    darks = [crop(matrix, rows, cols) for matrix in darks]
    master_dark_data = makeMasterDark(
        saveAs=CALIBRATION_OUTPUT_FOLDER / MASTER_DARK_NAME,
        headerToCopyFromName=next(
            get_darks(NIGHT_INPUT_CALIBRATION_FOLDER, image_duration)
        ).absolute(),
        listOfDarkData=darks,
    )
    logger.info("Created master dark")
    del darks  # Deleting to free memory as we don't use darks anymore

    master_flat_data = getdata(night["masterflat"])
    # Copy the masterflat provided to the calibration frames
    masterflat_path = Path(night["masterflat"])
    shutil.copy(masterflat_path, CALIBRATION_OUTPUT_FOLDER)
    logger.info("Using pre-provided masterflat")

    if raw_img_prefix := night.get("image_prefix"):
        raw_images: List[RawImageFile] = [
            RawImageFile(file.absolute())
            for file in get_all_fit_files(
                NIGHT_INPUT_IMAGES_FOLDER, image_duration, prefix=raw_img_prefix
            )
        ]
    else:
        raw_images: List[RawImageFile] = list(
            get_raw_images(NIGHT_INPUT_IMAGES_FOLDER, image_duration)
        )

    logger.info("Processing images")
    no_of_images_to_combine = config["processing"]["no_of_images_to_combine"]
    logger.info(f"Using no of images to combine: {no_of_images_to_combine}")
    logger.info(f"Radii of extraction: {radii_of_extraction}")

    # We now Calibrate/Crop/Align/Combine/Extract set of images in the size of no of combination
    # Note the subtle typing difference between no_of_combined_images and no_of_images_to_combine
    no_of_combined_images = len(raw_images) // no_of_images_to_combine

    # Create a file for storing alignment transformation
    alignment_stats_file_name = AlignmentStatsFile.generate_file_name(night_date)
    alignment_stats_file = AlignmentStatsFile(output / alignment_stats_file_name)
    alignment_stats_file.create_file_and_write_header()
    logger.info("Created alignment stats file")

    log_files_to_normalize: List[LogFileCombinedFile] = []
    for nth_combined_image in range(no_of_combined_images):
        try:
            align_combined_extract(
                config,
                night,
                output,
                night_date,
                nth_combined_image,
                raw_images,
                master_dark_data,
                master_flat_data,
                alignment_stats_file,
                image_duration,
                log_files_to_normalize,
            )
        except Exception as e:
            tb = traceback.format_exc()
            logger.error("Exception during alignment combination extraction")
            logger.error(e)
            logger.error(tb)
            return

    # Intranight + Internight Normalization
    try:
        normalization_helper(
            radii_of_extraction,
            reference_log_file,
            log_files_to_normalize,
            image_duration,
            night_date,
            color_ref_file_path,
            output,
            logfile_combined_reference_logfile,
            is_running_as_part_of_process=True,
        )
    except Exception as e:
        tb = traceback.format_exc()
        logger.error("Exception during normalization/sky_bg generation")
        logger.error(e)
        logger.debug(tb)
        return


def start_data_processing_auxiliary(config: Config):
    """
    This function processes (one or more) nights defined in config dict by
    putting together various functionalities like calibration, alignment,
    extraction, and normalization together.
    """

    OUTPUT_PATH: Path = config["output"]["path"]
    # If directory doesn't exist create directory including necessary parent directories.
    OUTPUT_PATH.mkdir(parents=True, exist_ok=True)

    def process_nights_mapper(night):
        night_path: Path = night["path"]
        night_date = get_date_from_input_night_folder_name(night_path.name)
        OUTPUT_NIGHT_FOLDER = OUTPUT_PATH / get_output_folder_name_from_night_date(night_date)
        # Create output folder for the night, if it doesn't already exist
        OUTPUT_NIGHT_FOLDER.mkdir(exist_ok=True)
        process_night(night, config, OUTPUT_NIGHT_FOLDER, night_date)

    cpu_fraction = config["processing"]["cpu_fraction"]
    if cpu_fraction > 0:
        cpu_count = int(os.cpu_count() * cpu_fraction)
        print(f"Multiprocessing module used. CPU count: {cpu_count}")
        with mp.Pool(cpu_count) as p:  # Use 75% CPU
            p.map(process_nights_mapper, config["input"]["nights"])
    else:
        # Dont use multiprocessing
        for night in config["input"]["nights"]:
            print("Using single processor.")
            process_nights_mapper(night)


def start_data_processing(file_path: str):
    """
    Starts data processing with the configuration file `file_path` provided as the argument.
    Calls auxiliary function `start_data_processing_auxiliary` if the configuration is valid.
    """
    validate_file(Path(file_path), on_success=start_data_processing_auxiliary)



def surroundWith(npArray, size, value=0):
    """
    @param npArray: numpy array
    @param size: integer specifying the length of the edge
    @param value: (optional) value to surround with, 0 default

    Mutates the original array provided, npArray
    """
    # columns
    npArray[:, :size] = value
    npArray[:, -size:] = value

    # rows
    npArray[:size, :] = value
    npArray[-size:, :] = value



import logging
import traceback
from pathlib import Path
from typing import Callable, Dict, List, TypedDict

import numpy as np
from scipy.optimize import curve_fit

from m23.charts import draw_internight_brightness_chart, draw_internight_color_chart
from m23.constants import COLOR_NORMALIZED_FOLDER_NAME, FLUX_LOGS_COMBINED_FOLDER_NAME
from m23.exceptions import InternightException
from m23.file.color_normalized_file import ColorNormalizedFile
from m23.file.flux_log_combined_file import FluxLogCombinedFile
from m23.file.log_file_combined_file import LogFileCombinedFile
from m23.file.ri_color_file import RIColorFile
from m23.utils import (
    customMedian,
    get_date_from_input_night_folder_name,
    get_radius_folder_name,
)
from m23.utils.flux_to_magnitude import flux_to_magnitude

# Note that this code is implemented based on the internight normalization in IDL
# https://github.com/LutherAstrophysics/idl-files/blob/39dfa1c0c6d03d64020c42583bbcaa94655d69cc/inter_night_normalization_345.pro


class ConfigImage(TypedDict):
    rows: int


def internight_normalize(
    night: Path,
    logfile_combined_reference_file: LogFileCombinedFile,
    color_file: Path,
    radii_of_extraction: List[int],
) -> Dict[int, Dict[str, Dict[int, float]]]:
    """
    This function normalizes the Flux Logs Combined for a night with respect to
    the data in the reference night. It also saves the result of inter-night
    normalization.  We typically the image 71 on Aug 4 2003 night as the
    reference.

    Note that since the a night can have Flux Logs Combined for multiple radii
    of extraction, this functions creates a color normalized output folder for
    as many radii of extraction as specified. Note that for each specified
    radius of extraction specified, respective Flux Logs Combined folder has to
    exist in `night` Path provided.

    param: night: Night folder that contains Flux Logs Combined folder param:
    reference_file: The path to the reference file to use.  param: color_file:
    The path to the the mean R-I file to use.

    return: None

    Side effects:

    This function creates a 'Color Normalized' folder inside `night` Path and
    folders for each radius specified in radii_of_extraction inside 'Color
    Normalized' folder that contains a txt file of the inter-night normalized
    data for each star for that night.

    This function also logs to the default log file in `night`. See
    `process_nights` inside `processor` module for the details of the log file.


    Preconditions:
    `night` is a valid path containing Flux Logs Combined for all radius
    specified in `radii_of_extraction`

    `reference_file` is a valid file path in conventional reference file format

    `color_file` is a valid file path in conventional R-I color file format
    """
    normfactors = {}
    for radius in radii_of_extraction:
        normfactors[radius] = internight_normalize_auxiliary(
            night, logfile_combined_reference_file, color_file, radius
        )
    return normfactors


def internight_normalize_auxiliary(  # noqa
    night: Path,
    logfile_combined_reference_file: LogFileCombinedFile,
    color_file: Path,
    radius_of_extraction: int,
) -> Dict[str, Dict[int, float]]:
    """
    This is an auxiliary function for internight_normalize that's different from
    the `internight_normalize` because this function takes
    `radius_of_extraction` unlike `internight_normalize` that takes
    `radii_of_extraction`.

    See https://drive.google.com/file/d/1R1Xc9RhPEYXgF5jlmHvtmDqvrVWs6xfK/view?usp=sharing
    for explanation of this algorithm by Prof. Wilkerson.
    """
    # Setup logging
    night_date = get_date_from_input_night_folder_name(night)
    logger = logging.getLogger("LOGGER_" + str(night_date))
    logger.info(f"Running internight color normalization for {radius_of_extraction}px")

    # Flux logs for a particular radius for that night is our primary input to
    # this algorithm We are essentially calculating median values of the flux
    # logs combined for a star and multiplying it by a normalization factor. We
    # do this for each star.  How we calculate normalization factor is described
    # later below.
    FLUX_LOGS_COMBINED_FOLDER = (
        night / FLUX_LOGS_COMBINED_FOLDER_NAME / get_radius_folder_name(radius_of_extraction)
    )
    flux_logs_files: List[FluxLogCombinedFile] = [
        FluxLogCombinedFile(file) for file in FLUX_LOGS_COMBINED_FOLDER.glob("*")
    ]
    # Filter out the files that don't match conventional flux log combined file format
    flux_logs_files = list(filter(lambda x: x.is_valid_file_name(), flux_logs_files))

    color_data_file = RIColorFile(color_file)

    # Messy hackery used instead of just reading the attendance float value
    # to mimick IDL
    is_star_attendance_over_half: Dict[int, bool] = {}

    data_dict: ColorNormalizedFile.Data_Dict_Type = {}

    for log_file in flux_logs_files:
        # This dictionary holds the data for each
        # Star's median ADU, normalization factor and normalized ADU
        star_number = log_file.star_number()
        data_dict[star_number] = ColorNormalizedFile.StarData(
            # Median flux value
            log_file.specialized_median_for_internight_normalization(),
            np.nan,  # Normalized median
            np.nan,  # Norm factor
            # Star color in color ref file
            color_data_file.get_star_color(star_number),
            np.nan,  # Actual color value used
            log_file.attendance(),  # Attendance of the star for the night
            logfile_combined_reference_file.get_star_data(star_number).radii_adu[
                radius_of_extraction
            ]
            or np.nan,
        )
        is_star_attendance_over_half[star_number] = log_file.is_attendance_over_half()

    last_star_no = 2510

    # We calculate the ratio of signal in reference file data and the special
    # median signal for a star for the night. We do this for each stars with
    # >50% attendance.
    stars_signal_ratio: Dict[int, float] = {}
    # Note +1 because of python the way range works
    for star_no in range(1, last_star_no + 1):
        star_data = data_dict[star_no]
        # Only calculate the ratio for stars with >= 50% attendance for the night
        if is_star_attendance_over_half[star_no] and not np.isnan(star_data.median_flux):
            # if star_data.attendance >= 0.5:
            # Only include this star if it has a non-zero median flux
            ratio = star_data.reference_log_adu / star_data.median_flux
            stars_signal_ratio[star_no] = ratio

    all_ref_night = []
    for star in range(1, last_star_no + 1):
        all_ref_night.append(stars_signal_ratio.get(star, 0))

    # Now we try to find correction factors (aka. normalization factor) for
    # stars with mean r-i data. How we deal with stars without mean r-i data is
    # described later.  For those that have mean r-i color data,  we sort these
    # stars into 3 populations based on mean r-i,as they appear different on a
    # histogram of mean r-i and graphs of mean r-i vs.  ref/night.  Potentially,
    # they could be disk stars in the luster, disk stars in the field, and bulge
    # stars in the field. That remains to be looked into.  Next, we remove
    # outliers using an initial polynomial curve fit. Then we fit a final curve
    # to each section. At the end of this program, we use the curve fits to
    # normalize these stars with color data.

    # The following dictionary holds the information for which one of the three
    # regions (described above) do stars that have mean r-i values as well as
    # are more than 50% attendant on the night fall into. So we look at the
    # stars from color dict and ensure that it's also in the dictionary
    # stars_signal_ratio)

    # Map from star number to color_section number (either 1 or 2 or 3)
    stars_color_section_number: Dict[int, int] = {}
    for star_no in range(1, last_star_no + 1):
        star_color_value = data_dict[star_no].measured_mean_r_i
        if star_color_value > 0.135 and star_color_value <= 0.455:
            stars_color_section_number[star_no] = 1
        elif star_color_value > 0.455 and star_color_value <= 1.063:
            stars_color_section_number[star_no] = 2
        elif star_color_value > 1.063 and star_color_value <= 7:
            stars_color_section_number[star_no] = 3

    # Otherwise we don't include the star in the color_section number
    # We now remove outliers points from signal-ratio vs r-i graph.
    # Essentially, this program works by taking a 3rd degree polynomial curve
    # fit for each of the 3 intervals. Then, it calculates how far each point is
    # from its predicted point on the curve, the point - point on the curve.  It
    # bins and creates a histogram for these differences, and fits a Gaussian to
    # it (this is statistically useful in limiting the sway of outliers on the
    # mean and standard deviation). Then, any point more than 2.5 standard
    # deviations from the curve fit is removed as an outlier.

    # Here we loop over each section of the polynomial fit
    # This dict holds data for each section that we use outside the loop
    section_data = {}
    sections = [1, 2, 3]
    for section_number in sections:
        stars_to_include = []
        for star_no in stars_color_section_number:
            if (
                stars_color_section_number[star_no] == section_number
                and star_no in stars_signal_ratio
            ):
                stars_to_include.append(star_no)

        # Colors
        x_values = [data_dict[star_no].measured_mean_r_i for star_no in stars_to_include]
        # Signal ratios
        y_values = [stars_signal_ratio[star_no] for star_no in stars_to_include]

        # Sort by x
        x_sort_indices = np.argsort(x_values)
        x_values = np.array(x_values)[x_sort_indices]
        y_values = np.array(y_values)[x_sort_indices]
        stars_to_include = np.array(stars_to_include)[x_sort_indices]

        # These lines of code check to make sure the first or last data point in
        # the signal value isn't an outlier.  First or last data points tend to
        # greatly affect the polynomial fit, so if the data points are 2
        # standard deviations away from the next closest point, they are
        # replaced with the mean of the next two closest points.
        std_signals = np.std(y_values, ddof=1)
        beginning_diff = abs(y_values[0] - y_values[1]) / std_signals
        ending_diff = abs(y_values[-1] - y_values[-2]) / std_signals
        if beginning_diff > 2 or ending_diff > 2:
            modified_y_value = y_values.copy()  # Note making copy is important
            if beginning_diff > 2:
                # Note python slicing excludes last element, IDL's includes
                modified_y_value[0] = np.mean(modified_y_value[1:4])
            if ending_diff > 2:
                # Note python slicing excludes last element, IDL's includes
                modified_y_value[-1] = np.mean(modified_y_value[-4:-1])
            a, b, c, d = np.polyfit(x_values, modified_y_value, 3)  # Degree 3

            # ax^3 + bx^2 + cx + d
            def polynomial_fit_fn(x):
                return a * x**3 + b * x**2 + c * x + d

        else:
            a, b, c, d = np.polyfit(x_values, y_values, 3)  # Degree 3

            # ax^3 + bx^2 + cx + d
            def polynomial_fit_fn(x):
                return a * x**3 + b * x**2 + c * x + d

        # This list stores the difference between actual signal value and the
        # value given by fitted curve
        y_differences = [
            polynomial_fit_fn(x) - y_values[index] for index, x in enumerate(x_values)
        ]

        # We store the data of each section to later in a dictionary indexed by
        # the section number
        section_data[section_number] = {
            "stars_to_include": stars_to_include,
            "x_values": x_values,
            "y_values": y_values,
            # This is the difference of actual y to fitted y
            "y_differences": y_differences,
        }

    # This holds the y_differences for three sections calculated in the loop above
    y_differences = []
    for section_number in sections:
        y_differences += section_data[section_number]["y_differences"]

    y_diff_std = np.std(y_differences, ddof=1)
    y_diff_mean = np.mean(y_differences)
    y_diff_min = y_diff_mean - 5 * y_diff_std
    y_diff_max = y_diff_mean + 5 * y_diff_std
    y_no_of_bins = 10  # We want to use 10 bins, like IDL code
    try:
        bin_frequencies, bins_edges = np.histogram(
            y_differences, range=[y_diff_min, y_diff_max], bins=y_no_of_bins
        )
    except ValueError:
        tb = traceback.format_exc()
        logger.error("Internight norm encountered exception in making histogram of residuals")
        logger.debug(tb)
        raise InternightException

    bins_mid_values = []
    for index, current_value in enumerate(bins_edges[:-1]):
        next_value = bins_edges[index + 1]
        bins_mid_values.append((current_value + next_value) / 2)
    fit_coefficients, _ = curve_fit(
        n_term_3_gauss_fit, bins_mid_values, bin_frequencies, maxfev=5000
    )
    mean, sigma = fit_coefficients[1], fit_coefficients[2]
    sigma = abs(sigma)  # Important since sigma given by our curve fit could be negative

    # Now we find stars for which y_difference is more than 2std away from mean
    stars_outside_threshold = []
    top_threshold = mean + 2.5 * sigma
    bottom_threshold = mean - 2.5 * sigma
    # We now create a list of stars that are outside the specified threshold
    for section_number in sections:
        section_y_differences = section_data[section_number]["y_differences"]
        section_stars = section_data[section_number]["stars_to_include"]
        for index, y_diff in enumerate(section_y_differences):
            if y_diff < bottom_threshold or y_diff > top_threshold:
                if sigma != 0:  # Guard for when fit fails and sigma is 0
                    star_no = section_stars[index]
                    stars_outside_threshold.append(star_no)

    # Now we do a second degree polynomial fit for the stars in sections that
    # aren't in `stars_outside_threshold` Note that we have to fit individual
    # curves for each section like we did above
    color_fit_functions = {}  # Stores the color fit function for each section
    # We save the x and the y so we can use this to make chart later
    section_x_values = {}
    section_y_values = {}
    for section_number in sections:
        # Note that we're excluding stars in `stars_outside_threshold` list
        stars_in_section = section_data[section_number]["stars_to_include"]
        stars_to_include = [
            star_no for star_no in stars_in_section if star_no not in stars_outside_threshold
        ]
        x_values = [data_dict[star_no].measured_mean_r_i for star_no in stars_to_include]  # Colors
        y_values = [stars_signal_ratio[star_no] for star_no in stars_to_include]  # Signal ratios
        section_x_values[section_number] = x_values
        section_y_values[section_number] = y_values

        a, b, c = np.polyfit(x_values, y_values, 2)  # Second degree fit
        # Nested lambda necessary to create a, b, c as local variables
        polynomial_fit_fn = (lambda a, b, c: lambda x: a * x**2 + b * x + c)(
            a, b, c
        )  # ax^2 + bx + c
        color_fit_functions[section_number] = polynomial_fit_fn

    normfactors_to_return = {}

    # Create and save the color chart
    normfactors_to_return["color"] = draw_internight_color_chart(
        night,
        radius_of_extraction,
        section_x_values,
        section_y_values,
        color_fit_functions,
    )

    stars_magnitudes = {
        star: flux_to_magnitude(data_dict[star].median_flux, radius_of_extraction)
        for star in range(1, last_star_no + 1)
    }
    star_magnitudes_section = {}  # Map from star no to star section no
    for star, star_magnitude in stars_magnitudes.items():
        if star_magnitude < 11:
            star_magnitudes_section[star] = 1
        elif 11 <= star_magnitude < 12.5:
            star_magnitudes_section[star] = 2
        else:
            star_magnitudes_section[star] = 3

    # These regions are divided based on stars magnitudes for
    # stars that have valid magnitude (& attendance)
    region_1_stars, region_2_stars, region_3_stars = [], [], []
    for star in stars_signal_ratio:
        star_magnitude = stars_magnitudes[star]
        if star_magnitude < 11:
            region_1_stars.append(star)
        elif 11 <= star_magnitude < 12.5:
            region_2_stars.append(star)
        else:
            region_3_stars.append(star)

    # FROM IDL version of this code "by looking at the graphs for multiple
    # nights, and fitted each regions differently...", we will use their fits
    magnitude_fit_fn = {}

    # Region 1
    # Magnitudes
    region_1_x = [stars_magnitudes[star] for star in region_1_stars]
    # Signal Ratio
    region_1_y = [stars_signal_ratio[star] for star in region_1_stars]
    coeffs_1 = np.polyfit(region_1_x, region_1_y, 1)  # Linear fit
    # ax + b
    magnitude_fit_fn[1] = lambda x: coeffs_1[0] * x + coeffs_1[1]

    # Region 2
    # Magnitudes
    region_2_x = [stars_magnitudes[star] for star in region_2_stars]
    # Signal Ratio
    region_2_y = [stars_signal_ratio[star] for star in region_2_stars]
    coeffs_2 = np.polyfit(region_2_x, region_2_y, 2)  # 2nd degree fit
    magnitude_fit_fn[2] = lambda x: (
        coeffs_2[0] * x**2 + coeffs_2[1] * x + coeffs_2[2]
    )  # ax^2 + bx + c

    # Region 3
    # Signal ratios
    region_3_x = [stars_magnitudes[star] for star in region_3_stars]
    region_3_y = [stars_signal_ratio[star] for star in region_3_stars]
    # For region 3, we just return the median of y values

    # magnitude_fit_fn[3] = lambda x: np.median(region_3_y)
    # We're using IDL like median here for getting results close to IDL's,
    # but np.median is better
    magnitude_fit_fn[3] = lambda x: customMedian(region_3_y)

    # Create and save the brightness chart
    normfactors_to_return["brightness"] = draw_internight_brightness_chart(
        night,
        radius_of_extraction,
        {1: region_1_x, 2: region_2_x, 3: region_3_x},
        {1: region_1_y, 2: region_2_y, 3: region_3_y},
        magnitude_fit_fn,
    )

    # Write normfactors for all stars
    for star_no in sorted(data_dict.keys()):
        star_data = data_dict[star_no]
        color = star_data.measured_mean_r_i
        norm_factor = None

        if 0.135 <= color < 7:
            color_section_number = stars_color_section_number[star_no]
            norm_factor = color_fit_functions[color_section_number](star_data.measured_mean_r_i)

        # Now for the stars that didn't have good R-I values which is (< 0.135
        # or >= 7) we calculate the normfactors based on the brightness fit
        else:
            # If the star is a known LPV that doesn't have a color, we calculate
            # the normfactor for it by providing a custom calculated color value
            special_star_value = get_normfactor_for_special_star(star_no, color_fit_functions[3])

            if special_star_value:
                # Note we're mutating color variable here
                norm_factor, color = special_star_value
            else:
                magnitude = stars_magnitudes[star_no]
                magnitude_section_no = star_magnitudes_section[star_no]
                norm_factor = magnitude_fit_fn[magnitude_section_no](magnitude)

        # Save the normfactor the star only if we have the normfactor for it

        if star_data.attendance < 0.5 or not star_data.median_flux > 0.001:
            normalized_median_flux = 0
        else:
            normalized_median_flux = norm_factor * star_data.median_flux
        # Replace is used because mutating a namedtuple directly isn't allowed
        data_dict[star_no] = star_data._replace(
            norm_factor=norm_factor,
            normalized_median_flux=normalized_median_flux,
            used_mean_r_i=color,
        )

    # Save data
    OUTPUT_FOLDER = (
        night / COLOR_NORMALIZED_FOLDER_NAME / get_radius_folder_name(radius_of_extraction)
    )
    output_file = OUTPUT_FOLDER / ColorNormalizedFile.get_file_name(
        night_date, radius_of_extraction
    )
    ColorNormalizedFile(output_file.absolute()).save_data(data_dict, night_date)

    # output_file = OUTPUT_FOLDER
    logger.info(f"Completed internight color normalization for {radius_of_extraction}px")
    return normfactors_to_return


def get_normfactor_for_special_star(
    star_no: int, fit_fn: Callable[[float], float]
) -> float | None:
    """
    For each of the special LPV stars that don't have a color value we calculate
    the normfactor by providing color values manually. For how we got these
    color values ask Prof. Wilkerson. Note that in the IDL code that this was
    implemented in python from the fit_fn was the polynomial fit function from
    the first region of the Signal Ratio vs Color fit. Note **not** brightness
    fit

    param: star_no : Star number
    param: fit_fn: a fit function that takes color values and returns the the
            fitted y

    return : The a tuple of normfactor for the star if the star is one of the
            special stars and the color value used Returns none if the star isn't a
            special star
    """
    stars_to_color_values = {
        814: 2.6137,
        1223: 3.6242,
        1654: 2.8866,
        1702: 2.9175,
        1716: 2.6137,
        1843: 2.7849,
        2437: 2.5545,
        2509: 2.7816,
        2510: 3.0923,
    }
    if star_no in stars_to_color_values:
        color = stars_to_color_values[star_no]
        return fit_fn(color), color


def n_term_3_gauss_fit(
    x,
    a0,
    a1,
    a2,
):
    z = (x - a1) / a2
    y = a0 * np.exp(-(z**2) / 2)
    return y



import datetime
import decimal
import math
from typing import Union

import ephem
import numpy as np

dec = decimal.Decimal

# Most of the code for Moon `position` and `phase` was taken from
# https://gist.github.com/miklb/ed145757971096565723


class MoonPositions:
    # Underscores instead of spaces are used for ease with csv processing
    NewMoon = "New_Moon"
    WaxingCrescent = "Waxing_Crescent"
    FirstQuarter = "First_Quarter"
    WaxingGibbous = "Waxing_Gibbous"
    FullMoon = "Full_Moon"
    WaningGibbous = "Waning_Gibbous"
    LastQuarter = "Last_Quarter"
    WaningCrescent = "Waning_Crescent"


def position(now=None):
    """
     Returns a decimal representation for the phase of the moon
    Takes into account waning and waxing. Full moon is about 0.5
    """
    if now is None:
        now = datetime.datetime.now()

    diff = now - datetime.datetime(2001, 1, 1)
    days = dec(diff.days) + (dec(diff.seconds) / dec(86400))
    lunations = dec("0.20439731") + (days * dec("0.03386319269"))

    return lunations % dec(1)


def phase(d: Union[datetime.date, datetime.datetime]):
    """
    Returns the phase (in str) of the moon in the given date
    """
    pos = position(d)
    index = (pos * dec(8)) + dec("0.5")
    index = math.floor(index)
    return {
        0: MoonPositions.NewMoon,
        1: MoonPositions.WaxingCrescent,
        2: MoonPositions.FirstQuarter,
        3: MoonPositions.WaxingGibbous,
        4: MoonPositions.FullMoon,
        5: MoonPositions.WaningGibbous,
        6: MoonPositions.LastQuarter,
        7: MoonPositions.WaningCrescent,
    }[int(index) & 7]


def get_moon_DE_and_RA(
    date: datetime.datetime,
):  # Returns the declination and right ascension of the moon for the given date and time in UTC
    # date parameter is given as a string 'year/month/day hh:mm:ss'

    observer = ephem.Observer()  # Decorah's latitude and longitude
    observer.lat = "43.3017"  # North is considered +ve
    observer.lon = "-91.79"  # West is considered -ve

    observer.date = date
    moon = ephem.Moon()

    moon.compute(observer)  # Gets the moon's RA and DE

    ra = moon.ra * 12 / ephem.pi  # Right ascension in hours
    dec = moon.dec * 180 / ephem.pi  # Declination in degrees

    return (dec, ra)


def moon_distance(
    date: datetime.datetime,
):
    """
    Returns the angle distance (in degrees) between the moon and our m23 cluster
    Note that the date is UTC date
    """

    # Note that DE is in degrees, RA is in hours
    # To convert RA to degrees, multiply by 15
    moon_DE, moon_RA = get_moon_DE_and_RA(date)

    # M23 declination and right ascension
    cluster_RA = 269.5667
    cluster_DE = -19.0186

    moon_alpha = moon_DE * np.pi / 180  # Moon declination in radians
    moon_beta = moon_RA * 15 * np.pi / 180  # Moon right ascension in radians

    cluster_alpha = cluster_DE * np.pi / 180  # Cluster declination in radians
    cluster_beta = cluster_RA * np.pi / 180  # Cluster right ascension in radians

    beta_difference = np.abs(moon_beta - cluster_beta)

    angle_cos = np.sin(moon_alpha) * np.sin(cluster_alpha) + np.cos(moon_alpha) * np.cos(
        cluster_alpha
    ) * np.cos(beta_difference)
    angle = np.arccos(angle_cos)  # Angle in radians between the two objects

    return angle * 180 / np.pi



from pathlib import Path
from typing import Tuple

import astroalign as ast
import numpy as np
import numpy.typing as npt
from astropy.io.fits import getdata as getfitsdata

from m23.exceptions import CouldNotAlignException

ScaleType = float
RotationType = float
TranslationXType = float
TranslationYType = float
AlignmentTransformationType = Tuple[RotationType, TranslationXType, TranslationYType, ScaleType]


def image_alignment(
    image_data_to_align: npt.NDArray, ref_image_name: str | Path
) -> Tuple[npt.NDArray, AlignmentTransformationType]:
    """
    Aligns the image data provided in `image_data_to_align` with respect to a reference image

    param: image_data_to_align: Numpy two dimensional array represents fits image
    param: ref_image_name: Pathlike object for filepath string to which the source data is
            to be aligned

    return:
        - Aligned image data
        - Transformation object which is a tuple of information about the alignment

    raises:
        ValueError: If it cannot find more than 3 stars on any input.
    """

    ref_image_data = getfitsdata(ref_image_name)

    # Note it's important to use dtype of float
    target_fixed = np.array(ref_image_data, dtype="float")
    source_fixed = np.array(image_data_to_align, dtype="float")

    try:
        t, __ = ast.find_transform(
            source=source_fixed,
            target=target_fixed,
            max_control_points=50,
            # Detection signal states how much sigma higher must the signal be w. bg
            detection_sigma=5,
            min_area=5,
        )
    except ast.MaxIterError:
        raise CouldNotAlignException
    except ValueError:
        raise CouldNotAlignException

    aligned_image_data, _ = ast.apply_transform(
        t,
        source_fixed,
        target_fixed,
        fill_value=0,
    )
    translation_x, translation_y = t.translation
    transformation_metrics: AlignmentTransformationType = (
        t.rotation,
        translation_x,
        translation_y,
        t.scale,
    )
    return aligned_image_data, transformation_metrics
